#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#define uint8_t unsigned char
#define uint16_t unsigned short
#define uint32_t unsigned int

char *file_image = NULL;
int file_length;

struct NE_header_struc {
  char sig[2];                 //"NE"
  uint8_t MajLinkerVersion;    //The major linker version
  uint8_t MinLinkerVersion;    //The minor linker version
  uint16_t EntryTableOffset;   //Offset of entry table, see below
  uint16_t EntryTableLength;   //Length of entry table in bytes
  uint32_t FileLoadCRC;        //UNKNOWN - PLEASE ADD INFO
  uint16_t ProgFlags;          //Program flags, bitmapped
  uint8_t AutoDataSegIndex;    //The automatic data segment index
  uint16_t InitHeapSize;       //The intial local heap size
  uint16_t InitStackSize;      //The inital stack size
  uint32_t EntryPoint;         //CS:IP entry point, CS is index into segment table
  uint32_t InitStack;          //SS:SP inital stack pointer, SS is index into segment table
  uint16_t SegCount;           //Number of segments in segment table
  uint16_t ModRefs;            //Number of module references (DLLs)
  uint16_t NoResNamesTabSiz;   //Size of non-resident names table, in bytes (Please clarify non-resident names table)
  uint16_t SegTableOffset;     //Offset of Segment table
  uint16_t ResTableOffset;     //Offset of resources table
  uint16_t ResidNamTable;      //Offset of resident names table
  uint16_t ModRefTable;        //Offset of module reference table
  uint16_t ImportNameTable;    //Offset of imported names table (array of counted strings, terminated with string of length 00h)
  uint32_t OffStartNonResTab;  //Offset from start of file to non-resident names table
  uint16_t MovEntryCount;      //Count of moveable entry point listed in entry table
  uint16_t FileAlnSzShftCnt;   //File allignment size shift count (0=9(default 512 byte pages))
  uint16_t nResTabEntries;     //Number of resource table entries
  uint8_t targOS;              //Target OS
  uint8_t OS2EXEFlags;         //Other OS/2 flags
  uint16_t retThunkOffset;     //Offset to return thunks or start of gangload area - what is gangload?
  uint16_t segrefthunksoff;    //Offset to segment reference thunks or size of gangload area
  uint16_t mincodeswap;        //Minimum code swap area size
  uint8_t expctwinver[2];      //Expected windows version (minor first)
};

struct NE_seg_struc {
  uint16_t seg_offset_in_file;
  uint16_t seg_len;
  uint16_t seg_flags;
  uint16_t seg_alloc_size;
};

struct NE_reloc_entry_struc
{
  uint8_t rel_src_type;
  uint8_t rel_flags;
  uint16_t rel_src_offset;
  uint16_t rel_param_1; // depends on rel_flags
  uint16_t rel_param_2; //
};

struct NE_header_struc *ne_header;
char *ne_image;

// -------------------------------------------------------------------------
int parse_ne_header(void)
{
  char *target_os;
  int ne_offset;

  char prog_name_buffer[1024];
  int prog_name_length;

  if (memcmp(file_image, "MZ", 2))
  {
    printf("Wrong exe file, no MZ header\n");
    return -11;
  }

  ne_offset = * (unsigned short *)(file_image + 0x3C);

  if (memcmp(file_image + ne_offset, "NE", 2))
  {
    printf("Wrong exe file, no NE header\n");
    return -12;
  }
  
  ne_header = (struct NE_header_struc *)(file_image + ne_offset);
  ne_image = file_image + ne_offset; // the same

  if ( (ne_header->targOS == 2) && (ne_header->ProgFlags & 0x800) )
  {
    printf("The file is self-loading.\n");
  }

  prog_name_length = *(file_image + ne_header->OffStartNonResTab);
  if (prog_name_length)
  {
    memset(prog_name_buffer, 0, sizeof(prog_name_buffer));
    memcpy(prog_name_buffer, file_image + ne_header->OffStartNonResTab + 1, prog_name_length);
    printf("Program title '%s'\n", prog_name_buffer);
  }


  switch ( ne_header->targOS )
  {
    case 0u:
      target_os = "unknown";
      break;
    case 1u:
      target_os = "OS/2";
      break;
    case 2u:
      target_os = "MS Windows";
      break;
    case 3u:
      target_os = "DOS 4.x";
      break;
    case 4u:
      target_os = "MS Windows 386";
      break;
    case 5u:
      target_os = "Borland Operating System Services (BOSS)";
      break;
    default:
      target_os = "unknown";
      break;
  }

  printf("Target operating system         %s (%04X)\n", target_os, ne_header->targOS);
  printf("File Load CRC                   0%08X\n", ne_header->FileLoadCRC);
  printf("Program Entry Point   (CS:IP)   %04X:%04X\n", ne_header->EntryPoint >> 16, ne_header->EntryPoint & 0xFFFF);
  printf("Initial Stack Pointer (SS:SP)   %04X:%04X\n", ne_header->InitStack >> 16, ne_header->InitStack & 0xFFFF);
  printf("Auto Data Segment Index         %04Xh  ( %5u. )\n", ne_header->AutoDataSegIndex, ne_header->AutoDataSegIndex);
  printf("Initial Local Heap Size         %04Xh  ( %5u. )\n", ne_header->InitHeapSize, ne_header->InitHeapSize);
  printf("Initial Stack Size              %04Xh  ( %5u. )\n", ne_header->InitStackSize, ne_header->InitStackSize);
  printf("Linker Version                  %d.%d\n", ne_header->MajLinkerVersion, ne_header->MinLinkerVersion);
  printf("Minimum code swap area size     %d\n", ne_header->mincodeswap);
  printf("Expected Windows Version        %d.%d\n", ne_header->expctwinver[1], ne_header->expctwinver[0]);
  printf("Program flags (%04X):\n", ne_header->ProgFlags);

  printf(!(ne_header->ProgFlags & 0x8000) ? "\tApplication\n" : "\tDLL\n");

  if (ne_header->ProgFlags & 2)
  {
    printf("\tMultiple data\n");
  } else
  {
    if (ne_header->ProgFlags & 1)
    {
      printf("\tSingle data\n");
    } else
    {
      printf("\tNo data\n");
    }
  }

  switch (ne_header->ProgFlags & 0x300)
  {
    case 0x100:
      printf("\tIncompatible with PM\n");
      break;

    case 0x200:
      printf("\tCompatible with PM\n");
      break;

    case 0x300:
      printf("\tUses windowing API\n");
      break;
  }

  if (ne_header->ProgFlags & 0x800)
  {
    switch(ne_header->targOS)
    {
      case 2:
        printf("\tSelf loading\n");
        break;

      case 1: 
        printf("\tFamily API\n");
        break;

      default:
        printf("\tUnknown (0x0800)\n");
    }
  }

  if (ne_header->ProgFlags & 0x2000)
  {
    printf("\tErrors in image\n");
  }

  printf("\nOther EXE Flags (%04X):\n", ne_header->OS2EXEFlags);

  if ( ne_header->OS2EXEFlags & 1 )
  {
    printf("\tLong file names\n");
  }

  if ( ne_header->OS2EXEFlags & 2 )
  {
    printf("\t2.X protected mode\n");
  }

  if ( ne_header->OS2EXEFlags & 4 )
  {
    printf("\t2.X proportional font\n");
  }

  if ( ne_header->OS2EXEFlags & 8 )
  {
    printf("\tFastload area (start: %08X, size: %08X)\n",
      ne_header->retThunkOffset << ne_header->FileAlnSzShftCnt,
      ne_header->segrefthunksoff << ne_header->FileAlnSzShftCnt);
  }

  return 0;
}


// -------------------------------------------------------------------------

struct rsrc_type_struc {
  uint16_t rsrc_type_id;
  uint16_t rsrc_type_number;
  uint32_t rsrc_type_reserved1;
};

struct rsrc_entry_struc {
  uint16_t rsrc_offset;
  uint16_t rsrc_length;
  uint16_t rsrc_flags;
  uint16_t rsrc_id;
  uint32_t rsrc_reserved1;
};


// -------------------------------------------------------------------------

int load_image(char *filename)
{
  FILE *input_file;
  struct stat sb;
  int result;

  if ((result = stat(filename, &sb)) < 0)
  {
    printf("**Error**: stat returned error\n");
    return -1;
  }
  
  file_length = sb.st_size;
  printf("File length is %d\n", file_length);

  if((file_image = (char *)malloc(file_length)) == NULL)
  {
    printf("**Error**: failed to allocate memory for file\n");
    return -2;
  }

  if ((input_file = fopen(filename, "rb")) == NULL)
  {
    printf("**Error**: Failed to open input file. Check permissions.\n");
    return -3;
  }

  if (fread(file_image, sizeof(char), file_length, input_file) != file_length)
  {
    printf("**Error**: Failed to read input file\n");
    return -4;
  }

  fclose(input_file);
  return 0;
}

// -------------------------------------------------------------------------
struct reloc_list_entry
{
//  struct NE_reloc_entry_struc raw_entry;
  struct reloc_list_entry *next;


  int rel_src_type;
  int rel_flags;
  int rel_src_offset;
  int rel_module_ordinal;
  int rel_proc_name_ptr;
  int rel_proc_ordinal;
  int rel_target_segment;
  int rel_target_offset;
  int rel_target_entry_ordinal;
  int rel_osfixup;
};

struct seg_list_entry
{
  char * seg_data;
  int seg_flags;
  int seg_ordinal;
  int seg_length;
  int seg_alloc_length;
  int seg_offset; // offset in file
  int seg_reloc_count;
  struct reloc_list_entry *seg_relocs;
  struct seg_list_entry *next;
};

struct seg_list_entry *seg_list_root;

int seg_count;

// segment flags
#define SEG_FLAGS_DATA             0x0001 // bit 0
#define SEG_FLAGS_CODE             0x0000 //
                                   
#define SEG_FLAGS_ALLOCATED        0x0002
#define SEG_FLAGS_LOADED           0x0004
#define SEG_FLAGS_ITERATED         0x0008
#define SEG_FLAGS_MOVEABLE         0x0010
#define SEG_FLAGS_PURE             0x0020
#define SEG_FLAGS_PRELOADED        0x0040
                                   
#define SEG_FLAGS_PROTECTED        0x0080
#define SEG_FLAGS_READ_ONLY        0x0001 // only if SEG_FLAGS_PROTECTED is enabled
#define SEG_FLAGS_EXEC_ONLY        0x0000 //
                                   
#define SEG_FLAGS_RELOCATIONS      0x0100 // of our interest
#define SEG_FLAGS_CONFORMING       0x0200
#define SEG_FLAGS_DISCARDABLE      0x1000
                                   
#define SEG_FLAGS_DPL0             (0 << 10)
#define SEG_FLAGS_DPL1             (1 << 10)
#define SEG_FLAGS_DPL2             (2 << 10)
#define SEG_FLAGS_DPL3             (3 << 10)
                                   
// relocation flags                                   
#define REL_SRC_MASK               0x0F
#define REL_SRC_LOBYTE             0x00
#define REL_SRC_SEGMENT            0x02
#define REL_SRC_FAR_PTR            0x03
#define REL_SRC_OFFSET             0x05
                                   
#define REL_FLAGS_TARGET_MASK      0x03
#define REL_FLAGS_INTERNALREF      0x00
#define REL_FLAGS_IMPORTORDINAL    0x01
#define REL_FLAGS_IMPORTNAME       0x02
#define REL_FLAGS_OSFIXUP          0x03
#define REL_FLAGS_ADDITIVE         0x04 // bit flag



// -------------------------------------------------------------------------
int free_segment_tree(void)
{
  struct seg_list_entry *curr_seg_list_node, *next_seg_list_node;
  struct reloc_list_entry *curr_reloc_list_node, *next_reloc_list_node;

  if (seg_list_root)
  {
    for (curr_seg_list_node = seg_list_root; curr_seg_list_node; curr_seg_list_node = next_seg_list_node)
    {
      if (curr_seg_list_node->seg_relocs)
      {
        for (curr_reloc_list_node = curr_seg_list_node->seg_relocs; curr_reloc_list_node; curr_reloc_list_node = next_reloc_list_node)
        {
          next_reloc_list_node = curr_reloc_list_node->next;
          free(curr_reloc_list_node);
        }

        curr_seg_list_node->seg_relocs = NULL;
      }

      next_seg_list_node = curr_seg_list_node->next;
      free(curr_seg_list_node);
    }
  
    seg_list_root = NULL;
  }

  return 0;
}

// -------------------------------------------------------------------------
int build_segment_tree(int build_relocs)
{
  int curr_segment, curr_reloc;
  struct NE_seg_struc *seg_entry;
  struct NE_reloc_entry_struc * reloc_entry;

  struct seg_list_entry *prev_seg_list_node, *curr_seg_list_node;
  struct reloc_list_entry *prev_reloc_list_node, *curr_reloc_list_node;

  printf("Number of segments: %d (%04X)\n", ne_header->SegCount, ne_header->SegCount);
  printf("Data page size: 2^%d (%04X)\n", ne_header->FileAlnSzShftCnt, 1 << ne_header->FileAlnSzShftCnt);

  seg_list_root = NULL;

  seg_count = ne_header->SegCount;

  for (curr_segment = 0; curr_segment < seg_count; curr_segment++)
  {
    if ((curr_seg_list_node = malloc(sizeof(struct seg_list_entry))) == NULL)
    {
      free_segment_tree();
      printf("Failed to allocate memory for segment tree item\n");
      return -1;
    } else
    {
      curr_seg_list_node->next = NULL;

      curr_seg_list_node->seg_relocs = NULL;

      seg_entry = (struct NE_seg_struc *)(ne_image + ne_header->SegTableOffset + curr_segment * sizeof(struct NE_seg_struc));

      curr_seg_list_node->seg_ordinal = curr_segment + 1;
      curr_seg_list_node->seg_flags = seg_entry->seg_flags;
      curr_seg_list_node->seg_offset = seg_entry->seg_offset_in_file << ne_header->FileAlnSzShftCnt;
      curr_seg_list_node->seg_data = file_image + curr_seg_list_node->seg_offset;
      curr_seg_list_node->seg_length = seg_entry->seg_len ? seg_entry->seg_len : 0x10000;
      curr_seg_list_node->seg_alloc_length = seg_entry->seg_alloc_size ? seg_entry->seg_alloc_size : 0x10000;

      if ((curr_seg_list_node->seg_flags & SEG_FLAGS_RELOCATIONS) && build_relocs)
      {
        curr_seg_list_node->seg_reloc_count = *((uint16_t*)(file_image 
          + (seg_entry->seg_offset_in_file << ne_header->FileAlnSzShftCnt) 
          + (seg_entry->seg_len ? seg_entry->seg_len : 0x10000)));
           
        for (curr_reloc = 0; curr_reloc < curr_seg_list_node->seg_reloc_count; curr_reloc ++)
        {
          if ((curr_reloc_list_node = malloc(sizeof(struct reloc_list_entry))) == NULL)
          {
            free_segment_tree();
            printf("Failed to allocate memory for relocation tree item\n");
            return -1;
          } else
          {
            curr_reloc_list_node->next = NULL; // to prevent free_segment_tree() fron traversing nonexistent tree
    
            reloc_entry = (struct NE_reloc_entry_struc *)(file_image 
              + (seg_entry->seg_offset_in_file << ne_header->FileAlnSzShftCnt) 
              + (seg_entry->seg_len ? seg_entry->seg_len : 0x10000)
              + sizeof(uint16_t) 
              + curr_reloc * sizeof(struct NE_reloc_entry_struc));
    
            memset(curr_reloc_list_node, 0, sizeof(struct reloc_list_entry));
    
//            curr_reloc_list_node->raw_entry = *reloc_entry;



            // decoding the raw part
            curr_reloc_list_node->rel_src_type = reloc_entry->rel_src_type;
            curr_reloc_list_node->rel_flags = reloc_entry->rel_flags;
            curr_reloc_list_node->rel_src_offset = reloc_entry->rel_src_offset;

            switch(curr_reloc_list_node->rel_flags & REL_FLAGS_TARGET_MASK)
            {
              case REL_FLAGS_INTERNALREF:
                curr_reloc_list_node->rel_target_segment = reloc_entry->rel_param_1 & 0xFF;

                if (curr_reloc_list_node->rel_target_segment == 0xFF)
                {
                  // movable segment
                  curr_reloc_list_node->rel_target_entry_ordinal = reloc_entry->rel_param_2;
                } else
                {
                  // fixed segment
                  curr_reloc_list_node->rel_target_offset = reloc_entry->rel_param_2;
                }
                break;

              case REL_FLAGS_IMPORTORDINAL:
                curr_reloc_list_node->rel_module_ordinal = reloc_entry->rel_param_1;
                curr_reloc_list_node->rel_proc_ordinal = reloc_entry->rel_param_2;
                break;

              case REL_FLAGS_IMPORTNAME:
                curr_reloc_list_node->rel_module_ordinal = reloc_entry->rel_param_1;
                curr_reloc_list_node->rel_proc_name_ptr = reloc_entry->rel_param_2;
                break;

              case REL_FLAGS_OSFIXUP:
                curr_reloc_list_node->rel_osfixup = reloc_entry->rel_param_1;
            }



    
            if (curr_seg_list_node->seg_relocs == NULL)
            {
              curr_seg_list_node->seg_relocs = curr_reloc_list_node;
            } else
            {
              prev_reloc_list_node->next = curr_reloc_list_node;
            }
            prev_reloc_list_node = curr_reloc_list_node;
          }
        }
      } else
      {
        curr_seg_list_node->seg_reloc_count = 0;
      }

      if (seg_list_root == NULL)
      {
        seg_list_root = curr_seg_list_node;
      } else
      {
        prev_seg_list_node->next = curr_seg_list_node;
      }
      prev_seg_list_node = curr_seg_list_node;

    }
  }

  return 0;
}

// -------------------------------------------------------------------------
int load_relocs_tree(int segment, char *filename)
{
  struct NE_reloc_entry_struc * reloc_entry;

  struct seg_list_entry *seg_list_node;
  struct reloc_list_entry *prev_reloc_list_node, *curr_reloc_list_node;

  struct stat sb;

  int result;
  int reloc_number;
  int seg_attr;
  int curr_reloc;
  int reloc_file_length;
  FILE *reloc_file;
  char *reloc_file_image;

  char *curr_ptr;
  int segment_found;

  if ((result = stat(filename, &sb)) < 0)
  {
    printf("**Error**: stat returned error\n");
    return -1;
  }
  
  reloc_file_length = sb.st_size;

  if((reloc_file_image = (char *)malloc(reloc_file_length)) == NULL)
  {
    printf("**Error**: failed to allocate memory for relocations file\n");
    return -2;
  }

  if ((reloc_file = fopen(filename, "rb")) == NULL)
  {
    printf("**Error**: Failed to open input file. Check permissions.\n");
    return -3;
  }

  if (fread(reloc_file_image, sizeof(char), reloc_file_length, reloc_file) != reloc_file_length)
  {
    printf("**Error**: Failed to read input file\n");
    return -4;
  }

  fclose(reloc_file);
 
  if (seg_list_root)
  {
    segment_found = 0;
    for (seg_list_node = seg_list_root; seg_list_node; seg_list_node = seg_list_node->next)
    {
      if (seg_list_node->seg_ordinal == segment)
      {
        segment_found = 1;
        break;
      }
    }

    if (! segment_found)
    {
      printf("**Error**: Required segment %02x not found\n", segment);
      return -5;
    }

    curr_ptr = reloc_file_image;

    reloc_number = *((uint16_t *)curr_ptr);
    curr_ptr += sizeof(uint16_t);

    seg_attr = *((uint16_t *)curr_ptr);
    curr_ptr += sizeof(uint16_t);
    seg_list_node->seg_flags = seg_attr;

    seg_list_node->seg_relocs = NULL;

    for (curr_reloc = 0; curr_reloc < reloc_number; curr_reloc ++)
    {
      reloc_entry = ((struct NE_reloc_entry_struc *)curr_ptr);
      curr_ptr += sizeof(struct NE_reloc_entry_struc);

      if ((curr_reloc_list_node = malloc(sizeof(struct reloc_list_entry))) == NULL)
      {
        free_segment_tree();
        printf("Failed to allocate memory for relocation tree item\n");
        return -1;
      } else
      {
        curr_reloc_list_node->next = NULL; // to prevent free_segment_tree() fron traversing nonexistent tree

//        curr_reloc_list_node->raw_entry = *reloc_entry;


        // decoding the raw part
        curr_reloc_list_node->rel_src_type = reloc_entry->rel_src_type;
        curr_reloc_list_node->rel_flags = reloc_entry->rel_flags;
        curr_reloc_list_node->rel_src_offset = reloc_entry->rel_src_offset;

        switch(curr_reloc_list_node->rel_flags & REL_FLAGS_TARGET_MASK)
        {
          case REL_FLAGS_INTERNALREF:
            curr_reloc_list_node->rel_target_segment = reloc_entry->rel_param_1 & 0xFF;

            if (curr_reloc_list_node->rel_target_segment == 0xFF)
            {
              // movable segment
              curr_reloc_list_node->rel_target_entry_ordinal = reloc_entry->rel_param_2;
            } else
            {
              // fixed segment
              curr_reloc_list_node->rel_target_offset = reloc_entry->rel_param_2;
            }
            break;

          case REL_FLAGS_IMPORTORDINAL:
            curr_reloc_list_node->rel_module_ordinal = reloc_entry->rel_param_1;
            curr_reloc_list_node->rel_proc_ordinal = reloc_entry->rel_param_2;
            break;

          case REL_FLAGS_IMPORTNAME:
            curr_reloc_list_node->rel_module_ordinal = reloc_entry->rel_param_1;
            curr_reloc_list_node->rel_proc_name_ptr = reloc_entry->rel_param_2;
            break;

          case REL_FLAGS_OSFIXUP:
            curr_reloc_list_node->rel_osfixup = reloc_entry->rel_param_1;
        }


    
        if (seg_list_node->seg_relocs == NULL)
        {
          seg_list_node->seg_relocs = curr_reloc_list_node;
        } else
        {
          prev_reloc_list_node->next = curr_reloc_list_node;
        }
        prev_reloc_list_node = curr_reloc_list_node;
      }
    }
  }

  return 0;
}

// -------------------------------------------------------------------------


int display_segment_tree(void)
{
  struct seg_list_entry *seg_list_node;
  struct reloc_list_entry *reloc_list_node;
  char module_name[1024];
  char procedure_name[1024];
  int module_name_offset;

  char *reloc_src_type_msg;
  int curr_reloc;

  if (seg_list_root != NULL)
  {
    for (seg_list_node = seg_list_root; seg_list_node; seg_list_node = seg_list_node->next)
    {
      printf("\nSegment %02X Offset %04X Length %05X\n", 
        seg_list_node->seg_ordinal, 
        seg_list_node->seg_offset, 
        seg_list_node->seg_length);

      curr_reloc = 0;

      for (reloc_list_node = seg_list_node->seg_relocs; reloc_list_node; reloc_list_node = reloc_list_node->next)
      {
        curr_reloc++;

        switch (reloc_list_node->rel_src_type)
        {
          case REL_SRC_LOBYTE:
            reloc_src_type_msg = "LOBYTE";
            break;

          case REL_SRC_SEGMENT:
            reloc_src_type_msg = "SEGMENT";
            break;

          case REL_SRC_FAR_PTR:
            reloc_src_type_msg = "FAR_PTR";
            break;

          case REL_SRC_OFFSET:
            reloc_src_type_msg = "OFFSET";
        }

        switch (reloc_list_node->rel_flags & REL_FLAGS_TARGET_MASK)
        {
          case REL_FLAGS_INTERNALREF: 
            if ((reloc_list_node->rel_target_segment & 0xFF) == 0xFF)
            {
              // movable
              printf("%d %s INTERNALREF at %04X Movable ordinal %04X", 
                curr_reloc, reloc_src_type_msg, reloc_list_node->rel_src_offset, 
                reloc_list_node->rel_target_entry_ordinal);
            } else
            {
              // fixed
              printf("%d %s INTERNALREF at %04X Fixed ptr SEG_%02X(%d):%04X", 
                curr_reloc, reloc_src_type_msg, reloc_list_node->rel_src_offset, 
                reloc_list_node->rel_target_segment, reloc_list_node->rel_target_segment, reloc_list_node->rel_target_offset);
            }

            break;

          case REL_FLAGS_IMPORTORDINAL:
            module_name_offset = *((uint16_t *)(ne_image + ne_header->ModRefTable + 
              (reloc_list_node->rel_module_ordinal - 1) * sizeof(uint16_t)));
            memset(module_name, 0, sizeof(module_name));
            memcpy(module_name, ne_image + ne_header->ImportNameTable + module_name_offset + 1, 
              *(ne_image + ne_header->ImportNameTable + module_name_offset));

            printf("%d %s IMPORTORDINAL at %04X %s:%d (%04X)", 
              curr_reloc, reloc_src_type_msg, reloc_list_node->rel_src_offset, 
              module_name, reloc_list_node->rel_proc_ordinal, reloc_list_node->rel_proc_ordinal);

            break;

          case 2: // IMPORTNAME
            module_name_offset = *((uint16_t *)(ne_image + ne_header->ModRefTable + 
              (reloc_list_node->rel_module_ordinal - 1) * sizeof(uint16_t)));
            memset(module_name, 0, sizeof(module_name));
            memcpy(module_name, ne_image + ne_header->ImportNameTable + module_name_offset + 1, 
              *(ne_image + ne_header->ImportNameTable + module_name_offset));

            memset(procedure_name, 0, sizeof(procedure_name));
            memcpy(procedure_name, ne_image + ne_header->ImportNameTable + reloc_list_node->rel_proc_name_ptr + 1,
              *(ne_image + ne_header->ImportNameTable + reloc_list_node->rel_proc_name_ptr));

            printf("%d %s IMPORTNAME at %04X %s:%s", 
              curr_reloc, reloc_src_type_msg, reloc_list_node->rel_src_offset, 
              module_name, procedure_name);

            break;

          case 3: // OSFIXUP
            printf("%d %s OSFIXUP at %04X value %04X", 
              curr_reloc, reloc_src_type_msg, reloc_list_node->rel_src_offset, 
              reloc_list_node->rel_osfixup);
            break;
        }

        if (reloc_list_node->rel_flags & REL_FLAGS_ADDITIVE)
        {
          printf(" Additive\n");
        } else
        {
          printf("\n");
        }
        
      }

    }
  }

  return 0;
}


// -------------------------------------------------------------------------
int export_segment(char *filename, int seg_ordinal)
{
  FILE *output_file;

  struct seg_list_entry *seg_list_node;

  if (seg_list_root != NULL)
  {
    for (seg_list_node = seg_list_root; seg_list_node; seg_list_node = seg_list_node->next)
    {
      if (seg_list_node->seg_ordinal == seg_ordinal)
      {
        if ((output_file = fopen(filename, "wb")) == NULL)
        {
          printf("Failed to open output file %s. Check permissions.\n", filename);
        }
        fwrite(seg_list_node->seg_data, sizeof(char), seg_list_node->seg_length, output_file);
        fclose(output_file);
        return 0;
      }
    }

    printf("Requested segment not found, exiting...\n");
    return 0;

  } else
  {
    printf("Segment list root is empty...\n");
    return -1;
  }
}

// -------------------------------------------------------------------------
int delete_reloc_src(int seg_ordinal, int offset)
{
  struct seg_list_entry *seg_list_node;
  struct reloc_list_entry *prev_reloc_list_node, *curr_reloc_list_node;

  if (seg_list_root != NULL)
  {
    for (seg_list_node = seg_list_root; seg_list_node; seg_list_node = seg_list_node->next)
    {
      if (seg_list_node->seg_ordinal == seg_ordinal)
      {
        if (seg_list_node->seg_relocs)
        {
          prev_reloc_list_node = NULL;
          for (curr_reloc_list_node = seg_list_node->seg_relocs;
            curr_reloc_list_node; curr_reloc_list_node = curr_reloc_list_node->next)
          {
            if (curr_reloc_list_node->rel_src_offset == offset)
            {
              // delete reloc
              if (prev_reloc_list_node)
              {
                prev_reloc_list_node->next = curr_reloc_list_node->next;
              } else
              {
                seg_list_node->seg_relocs = curr_reloc_list_node->next;
              }

              free(curr_reloc_list_node);
              printf ("Relocation at %02X:%04X is successfully deleted\n", seg_ordinal, offset);
              return 0;
            }

            prev_reloc_list_node = curr_reloc_list_node;
          }
        }

        printf ("Requested relocation at %02X:%04X not found\n", seg_ordinal, offset);
        return -1;
      }
    }

    printf("Requested segment not found, exiting...\n");
    return -1;

  } else
  {
    printf("Segment list root is empty...\n");
    return -1;
  }


}

// -------------------------------------------------------------------------



/*
  struct seg_list_entry
{
  char * seg_data;
  int seg_flags;
  int seg_ordinal;
  int seg_length;
  int seg_alloc_length;
  int seg_offset; // offset in file
  int seg_reloc_count;
  struct reloc_list_entry *seg_relocs;
  struct seg_list_entry *next;
};

*/

/*
struct rsrc_type_struc {
  uint16_t rsrc_type_id;
  uint16_t rsrc_type_number;
  uint32_t rsrc_type_reserved1;
};

struct rsrc_entry_struc {
  uint16_t rsrc_offset;
  uint16_t rsrc_length;
  uint16_t rsrc_flags;
  uint16_t rsrc_id;
  uint32_t rsrc_reserved1;
};
*/
struct literal_list_struc
{
  int ordinal;
  int offset;
  int references;
  char *literal_string;
  struct literal_list_struc *next;
};

struct rsrc_entry_list_struc
{
  int ordinal;
  int rsrc_length;
  int rsrc_offset;
  struct rsrc_entry_struc raw_entry;
  struct literal_list_struc *literal_entry;
  struct rsrc_entry_list_struc *next;
  char *rsrc_data;
};

struct rsrc_type_list_struc
{
  struct rsrc_type_struc raw_entry;
  struct rsrc_entry_list_struc *rsrc_entries;
  struct rsrc_type_list_struc *next;
};

struct rsrc_type_list_struc * rsrc_list_root = NULL;

int rsrc_alignment;
// -------------------------------------------------------------------------
int free_resource_tree(void)
{
  struct rsrc_type_list_struc *curr_rsrc_type_list_node, *next_rsrc_type_list_node;
  struct rsrc_entry_list_struc *curr_rsrc_entry_list_node, *next_rsrc_entry_list_node;

  struct literal_list_struc *curr_rsrc_literal_list_node, *next_rsrc_literal_list_node;

  if (rsrc_list_root)
  {
    for (curr_rsrc_type_list_node = rsrc_list_root; curr_rsrc_type_list_node; curr_rsrc_type_list_node = next_rsrc_type_list_node)
    {
      if (curr_rsrc_type_list_node->rsrc_entries)
      {
        for (curr_rsrc_entry_list_node = curr_rsrc_type_list_node->rsrc_entries; 
          curr_rsrc_entry_list_node; curr_rsrc_entry_list_node = next_rsrc_entry_list_node)
        {
          next_rsrc_entry_list_node = curr_rsrc_entry_list_node->next;
          free(curr_rsrc_entry_list_node);
        }

        curr_rsrc_entry_list_node = NULL;
      }
      next_rsrc_type_list_node = curr_rsrc_type_list_node->next;
      free(curr_rsrc_type_list_node);
    }

    rsrc_list_root = NULL;
  } 

  return 0;
}


// -------------------------------------------------------------------------
int build_resource_tree(void)
{
  char *curr_ptr;
  struct rsrc_type_struc * rsrc_type;
  struct rsrc_entry_struc * rsrc_entry;
  struct rsrc_type_list_struc *curr_rsrc_type_list_node, *prev_rsrc_type_list_node;
  struct rsrc_entry_list_struc *curr_rsrc_entry_list_node, *prev_rsrc_entry_list_node;
  struct literal_list_struc *curr_rsrc_literal_list_node, *prev_rsrc_literal_list_node;

  int literal_length;
  int current_res;
  int cont_rsrc_idx = 0;
  int i;
  
  rsrc_alignment = *((uint16_t*)(ne_image + ne_header->ResTableOffset));

  curr_ptr = ne_image + ne_header->ResTableOffset + sizeof(uint16_t);

  rsrc_list_root = NULL;

  while (1)
  {
    rsrc_type = (struct rsrc_type_struc *) curr_ptr;
    if (! rsrc_type->rsrc_type_id)
    {
      break;
    } else
    {
      if ((curr_rsrc_type_list_node = malloc(sizeof(struct rsrc_type_list_struc))) == NULL)
      {
        free_resource_tree();
        printf("Failed to allocate memory for resource type tree item\n");
        return -1;
      } else
      {
        curr_rsrc_type_list_node->next = NULL;
        rsrc_type = (struct rsrc_type_struc *) curr_ptr;
        curr_rsrc_type_list_node->raw_entry = *rsrc_type;

        curr_ptr += sizeof(struct rsrc_type_struc);

        curr_rsrc_type_list_node->rsrc_entries = NULL;
        for (current_res = 0; current_res < rsrc_type->rsrc_type_number; current_res ++)
        {
          if ((curr_rsrc_entry_list_node = malloc(sizeof(struct rsrc_entry_list_struc))) == NULL)
          {
            free_resource_tree();
            printf("Failed to allocate memory for resource entry tree item\n");
            return -1;
          } else
          {
            cont_rsrc_idx ++;
            curr_rsrc_entry_list_node->ordinal = cont_rsrc_idx;

            curr_rsrc_entry_list_node->next = NULL;
            rsrc_entry = (struct rsrc_entry_struc *) curr_ptr;
            curr_rsrc_entry_list_node->raw_entry = *rsrc_entry;

            curr_rsrc_entry_list_node->rsrc_length = rsrc_entry->rsrc_length << rsrc_alignment;
            curr_rsrc_entry_list_node->rsrc_offset = ((rsrc_entry->rsrc_offset) << rsrc_alignment);
            curr_rsrc_entry_list_node->rsrc_data = file_image + ((rsrc_entry->rsrc_offset) << rsrc_alignment);

            if (rsrc_entry->rsrc_id & 0x8000)
            {
              curr_rsrc_entry_list_node->literal_entry = NULL;
            } else
            {
              if ((curr_rsrc_literal_list_node = malloc(sizeof(struct literal_list_struc))) == NULL)
              {
                free_resource_tree();
                printf("Failed to allocate memory for resource literal tree item\n");
                return -1;
              } else
              {
                curr_rsrc_literal_list_node->next = NULL;

                literal_length = *(ne_image + ne_header->ResTableOffset + rsrc_entry->rsrc_id);
                if ((curr_rsrc_literal_list_node->literal_string = malloc(literal_length + 1)) == NULL)
                {
                  free_resource_tree();
                  printf("Failed to allocate memory for resource literal string\n");
                  return -1;
                } else
                {
                  memcpy(curr_rsrc_literal_list_node->literal_string, ne_image + ne_header->ResTableOffset + rsrc_entry->rsrc_id + 1, literal_length);
                  curr_rsrc_literal_list_node->literal_string[literal_length] = 0;
                  curr_rsrc_literal_list_node->offset = rsrc_entry->rsrc_id;

                  curr_rsrc_entry_list_node->literal_entry = curr_rsrc_literal_list_node;
                }
              }
            }

            if (curr_rsrc_type_list_node->rsrc_entries == NULL)
            {
              curr_rsrc_type_list_node->rsrc_entries = curr_rsrc_entry_list_node;
            } else
            {
              prev_rsrc_entry_list_node->next = curr_rsrc_entry_list_node;
            }

            prev_rsrc_entry_list_node = curr_rsrc_entry_list_node;
          }
          curr_ptr += sizeof(struct rsrc_entry_struc);
        }


        if (rsrc_list_root == NULL)
        {
          rsrc_list_root = curr_rsrc_type_list_node;
        } else
        {
          prev_rsrc_type_list_node->next = curr_rsrc_type_list_node;
        }

        prev_rsrc_type_list_node = curr_rsrc_type_list_node;
      }
    }
  }

  return 0;

}


// -------------------------------------------------------------------------
int delete_resource_by_literal(char *literal)
{
  struct rsrc_type_list_struc *curr_rsrc_type_list_node, *prev_rsrc_type_list_node;
  struct rsrc_entry_list_struc *curr_rsrc_entry_list_node, *prev_rsrc_entry_list_node;

  if (rsrc_list_root)
  {
    prev_rsrc_type_list_node = NULL;

    for(curr_rsrc_type_list_node = rsrc_list_root; curr_rsrc_type_list_node; curr_rsrc_type_list_node = curr_rsrc_type_list_node->next)
    {
      prev_rsrc_entry_list_node = NULL;

      for (curr_rsrc_entry_list_node = curr_rsrc_type_list_node->rsrc_entries; 
        curr_rsrc_entry_list_node; curr_rsrc_entry_list_node = curr_rsrc_entry_list_node->next)
      {
        if (curr_rsrc_entry_list_node->literal_entry)
        {
          if (! strcmp(curr_rsrc_entry_list_node->literal_entry->literal_string, literal))
          {
            goto literal_found; // how to break two `for' loops at once?
          }
        }

        if (curr_rsrc_entry_list_node != curr_rsrc_type_list_node->rsrc_entries)
        {
          prev_rsrc_entry_list_node = curr_rsrc_entry_list_node;
        }
      }

      if (curr_rsrc_type_list_node != rsrc_list_root)
      {
        prev_rsrc_type_list_node = curr_rsrc_type_list_node;
      }
    }

    printf("Requested resource literal not found, nothing to do\n");
    return -1;

    literal_found:
    if (prev_rsrc_entry_list_node)
    {
      prev_rsrc_entry_list_node->next = curr_rsrc_entry_list_node->next; // remove current entry from the chain
    } else
    {
      curr_rsrc_type_list_node->rsrc_entries = curr_rsrc_entry_list_node->next; // remove the first entry
    }

    free(curr_rsrc_entry_list_node->literal_entry->literal_string);
    free(curr_rsrc_entry_list_node->literal_entry);
    free(curr_rsrc_entry_list_node);

    if (curr_rsrc_type_list_node->rsrc_entries == NULL) // check if empty type list node
    {
      if (prev_rsrc_type_list_node)
      {
        // not the first entry
        prev_rsrc_type_list_node->next = curr_rsrc_type_list_node->next;
      } else
      {
        // the first entry
        rsrc_list_root = curr_rsrc_type_list_node->next;
      }

      free(curr_rsrc_type_list_node);
    }
  } else
  {
    printf("Resource list root is empty, exiting\n");
    return -1;
  }

  return 0;
}

// -------------------------------------------------------------------------

int display_resource_tree(void)
{
  char *rsrc_type_msg;

  struct rsrc_type_list_struc *rsrc_type_list_node;
  struct rsrc_entry_list_struc *rsrc_entry_list_node;

  int literal_index;

  if (rsrc_list_root)
  {
    for (rsrc_type_list_node = rsrc_list_root; rsrc_type_list_node; rsrc_type_list_node = rsrc_type_list_node->next)
    {
      if ( rsrc_type_list_node->raw_entry.rsrc_type_id & 0x8000 )
      {
        switch ( rsrc_type_list_node->raw_entry.rsrc_type_id & 0x7FFF )
        {
          case 1:
            rsrc_type_msg = "Cursor";
            break;
          case 2:
            rsrc_type_msg = "Bitmap";
            break;
          case 3:
            rsrc_type_msg = "Icon";
            break;
          case 4:
            rsrc_type_msg = "Menu";
            break;
          case 5:
            rsrc_type_msg = "Dialog box";
            break;
          case 6:
            rsrc_type_msg = "String table";
            break;
          case 7:
            rsrc_type_msg = "Font directory";
            break;
          case 8:
            rsrc_type_msg = "Font component";
            break;
          case 9:
            rsrc_type_msg = "Accelerator table";
            break;
          case 10:
            rsrc_type_msg = "Resource data";
            break;
          case 12:
            rsrc_type_msg = "Cursor directory";
            break;
          case 14:
            rsrc_type_msg = "Icon directory";
            break;
          case 15:
            rsrc_type_msg = "Name table";
            break;
          case 16:
            rsrc_type_msg = "Version info";
            break;
          default:
            rsrc_type_msg = "Unknown";
            break;
        }
      } else
      {
        rsrc_type_msg = "User defined";
      }
    
      printf("Resource type %04X number %04X: %s\n",
        rsrc_type_list_node->raw_entry.rsrc_type_id, 
        rsrc_type_list_node->raw_entry.rsrc_type_number,
        rsrc_type_msg);
    
      if (rsrc_type_list_node->rsrc_entries)
      {
        for (rsrc_entry_list_node = rsrc_type_list_node->rsrc_entries; 
          rsrc_entry_list_node; rsrc_entry_list_node = rsrc_entry_list_node->next)
        {
          if (rsrc_entry_list_node->raw_entry.rsrc_id & 0x8000)
          {
            printf("\tOffset %08X length %08X flags %04X ID %04X ", 
              rsrc_entry_list_node->raw_entry.rsrc_offset << rsrc_alignment,
              rsrc_entry_list_node->rsrc_length,
              rsrc_entry_list_node->raw_entry.rsrc_flags,
              rsrc_entry_list_node->raw_entry.rsrc_id & 0x7FFF);
          } else
          {
            printf("\tOffset %08X length %08X flags %04X ID %s", 
              rsrc_entry_list_node->raw_entry.rsrc_offset << rsrc_alignment,
              rsrc_entry_list_node->rsrc_length,
              rsrc_entry_list_node->raw_entry.rsrc_flags,
              rsrc_entry_list_node->literal_entry->literal_string);
          }
    
          if (rsrc_entry_list_node->raw_entry.rsrc_flags & 0x10) printf(" Movable");
          if (rsrc_entry_list_node->raw_entry.rsrc_flags & 0x20) printf(" Pure");
          if (rsrc_entry_list_node->raw_entry.rsrc_flags & 0x40) printf(" Preloaded");
          printf("\n");
        }
    
      } else
      {
        printf("Empty type...\n");
      } 
    }
  }

  return 0;
}

// -------------------------------------------------------------------------
struct literal_list_struc *res_name_list_root = NULL;

// -------------------------------------------------------------------------
int build_resident_name_tree(void)
{
  char *curr_ptr;
  int literal_length;

  struct literal_list_struc *curr_res_name_list_node, *prev_res_name_list_node;

  res_name_list_root = NULL;

  for (curr_ptr = ne_image + ne_header->ResidNamTable; curr_ptr < ne_image + ne_header->ModRefTable; )
  {
    literal_length = *((uint8_t *)curr_ptr);

    if (literal_length == 0) break;

    if ((curr_res_name_list_node = malloc(sizeof(struct literal_list_struc))) == NULL)
    {
      printf("Failed to allocate memory for resident name list item\n");
      return -1;
    } else
    {
      curr_res_name_list_node->next = 0;

      curr_ptr ++;

      if ((curr_res_name_list_node->literal_string = malloc(literal_length + 1)) == NULL)
      {
        printf("Failed to allocate memory for resident name literal string\n");
        return -1;
      } else
      {
        memcpy(curr_res_name_list_node->literal_string, curr_ptr, literal_length);
        curr_res_name_list_node->literal_string[literal_length] = 0;
      }

      curr_ptr += literal_length;
      curr_res_name_list_node->ordinal = *((uint16_t *)curr_ptr);
      curr_ptr += sizeof(uint16_t);

      // update pointers
      if (res_name_list_root == NULL)
      {
        res_name_list_root = curr_res_name_list_node;
      } else
      {
        prev_res_name_list_node->next = curr_res_name_list_node;
      }
      prev_res_name_list_node = curr_res_name_list_node;
    }
  }

  return 0;
}

// -------------------------------------------------------------------------
int display_resident_name_tree(void)
{
  struct literal_list_struc *res_name_list_node;

  if (res_name_list_root)  
  {
    printf("\nResident names by ordinal:\n");
    for (res_name_list_node = res_name_list_root; res_name_list_node; res_name_list_node = res_name_list_node->next)
    {
      printf("\t%s\t%04X\n", res_name_list_node->literal_string, res_name_list_node->ordinal);
    }
  }

  return 0;
}

// -------------------------------------------------------------------------
struct literal_list_struc *module_list_root = NULL;

// -------------------------------------------------------------------------
int build_module_tree(void)
{
  int i;
  int literal_length;
  int ref;

  struct literal_list_struc *curr_module_list_node, *prev_module_list_node;

  module_list_root = NULL;

  for (i = 0; i < ne_header->ModRefs; i++)
  {
    if ((curr_module_list_node = malloc(sizeof(struct literal_list_struc))) == NULL)
    {
      printf("Failed to allocate memory for module list item\n");
      return -1;
    } else
    {
      curr_module_list_node->next = NULL;

      curr_module_list_node->ordinal = i + 1;
      ref = *((uint16_t *) (ne_image + ne_header->ModRefTable + i * sizeof(uint16_t)));
      curr_module_list_node->offset = ref;
      literal_length = *(ne_image + ne_header->ImportNameTable + ref);

      if ((curr_module_list_node->literal_string = malloc(literal_length + 1)) == NULL)
      {
        printf("Failed to allocate memory for module literal string\n");
        return -1;
      } else
      {
        memcpy(curr_module_list_node->literal_string, ne_image + ne_header->ImportNameTable + ref + 1, literal_length);
        curr_module_list_node->literal_string[literal_length] = 0;
      }

      // update pointers
      if (module_list_root == NULL)
      {
        module_list_root = curr_module_list_node;
      } else
      {
        prev_module_list_node->next = curr_module_list_node;
      }
      prev_module_list_node = curr_module_list_node;
    
    }
  }
}

// -------------------------------------------------------------------------
int display_module_tree(void)
{
  
  struct literal_list_struc *module_list_node;

  if (module_list_root)
  {
    printf("\nModules by ordinal:\n");
    for (module_list_node = module_list_root; module_list_node; module_list_node = module_list_node->next)
    {
      printf("\t%04X, offset %04x\t%s\n", 
        module_list_node->ordinal,
        module_list_node->offset,
        module_list_node->literal_string);
    }
  }

  return 0;
}


// -------------------------------------------------------------------------
struct entry_list_struc
{
  int flags;
  int segment;
  int offset;
  int ordinal;
  int size;
  struct entry_list_struc *next;
};

struct entry_bundle_list_struc
{
  int segment_indicator;
  int entry_count;
  int bundle_index;
  struct entry_list_struc *entries;
  struct entry_bundle_list_struc *next;
};

struct entry_bundle_list_struc * entry_list_root = NULL;
int entries = 0;

// -------------------------------------------------------------------------
int build_entry_tree(void)
{
  struct entry_bundle_list_struc *curr_entry_bundle_list_node, *prev_entry_bundle_list_node;
  struct entry_list_struc *curr_entry_list_node, *prev_entry_list_node;

  unsigned char *curr_ptr;
  int local_entries;
  int curr_entry;
  int cont_entry_idx = 0;
  int bundle_index = 0;

  curr_ptr = ne_image + ne_header->EntryTableOffset;

  while (-1)
  {
    local_entries = *((uint8_t *)curr_ptr);
    curr_ptr ++;

    if (! local_entries)
    {
      break;
    }

    if ((curr_entry_bundle_list_node = malloc(sizeof(struct entry_bundle_list_struc))) == NULL)
    {
      printf("Failed to allocate memory for entry bundle list item\n");
      return -1;
    } else
    {
      curr_entry_bundle_list_node->next = NULL;
      curr_entry_bundle_list_node->entries = NULL;
      curr_entry_bundle_list_node->entry_count = local_entries;

      bundle_index ++;
      curr_entry_bundle_list_node->bundle_index = bundle_index;

      curr_entry_bundle_list_node->segment_indicator = *((uint8_t *)curr_ptr);
      curr_ptr ++;

      if (curr_entry_bundle_list_node->segment_indicator)
      {
        for (curr_entry = 0; curr_entry < local_entries; curr_entry ++)
        {
          if ((curr_entry_list_node = malloc(sizeof(struct entry_list_struc))) == NULL)
          {
            printf("Failed to allocate memory for entry list item\n");
            return -1;
          } else
          {
            curr_entry_list_node->next = NULL;

            cont_entry_idx ++;
            curr_entry_list_node->ordinal = cont_entry_idx;

            if (curr_entry_bundle_list_node->segment_indicator == 0xFF)
            {
              // moveable, 6-byte entries
              curr_entry_list_node->flags = *((uint8_t *)curr_ptr);
              curr_entry_list_node->segment = *((uint8_t *)(curr_ptr + 3));
              curr_entry_list_node->offset = *((uint16_t *)(curr_ptr + 4));
              curr_entry_list_node->size = 6;

              curr_ptr += 6;
            } else
            {
              // fixed, 3-byte entries
              curr_entry_list_node->flags = *((uint8_t *)curr_ptr);
              curr_entry_list_node->segment = curr_entry_bundle_list_node->segment_indicator;
              curr_entry_list_node->offset = *((uint16_t *)(curr_ptr + 1));
              curr_entry_list_node->size = 3;

              curr_ptr += 3;
            }

            if (curr_entry_bundle_list_node->entries == NULL)
            {
              curr_entry_bundle_list_node->entries = curr_entry_list_node;
            } else
            {
              prev_entry_list_node->next = curr_entry_list_node;
            }

            prev_entry_list_node = curr_entry_list_node;
          }
        }
      } else
      {
        cont_entry_idx += local_entries;
      }

      if (entry_list_root == NULL)
      {
        entry_list_root = curr_entry_bundle_list_node;
      } else
      {
        prev_entry_bundle_list_node->next = curr_entry_bundle_list_node;
      }

      prev_entry_bundle_list_node = curr_entry_bundle_list_node;
    }
  }

  return 0;
}


// -------------------------------------------------------------------------
int display_entry_tree(void)
{
  struct entry_bundle_list_struc *entry_bundle_list_node;
  struct entry_list_struc *entry_list_node;

  if (entry_list_root)
  {
    for (entry_bundle_list_node = entry_list_root; 
      entry_bundle_list_node; entry_bundle_list_node = entry_bundle_list_node->next)
    {
      printf("Bundle %d Segment indicator %02X entries %d %s\n", 
        entry_bundle_list_node->bundle_index,
        entry_bundle_list_node->segment_indicator,
        entry_bundle_list_node->entry_count,
        (entry_bundle_list_node->segment_indicator == 0xFF) ? "Moveable" : "Static");

      if (entry_bundle_list_node->entries)
      {
        for (entry_list_node = entry_bundle_list_node->entries; entry_list_node; entry_list_node = entry_list_node->next)
        {
          printf("\tEntry %d ptr %02X:%04X flags %04X\n",
            entry_list_node->ordinal,
            entry_list_node->segment,
            entry_list_node->offset,
            entry_list_node->flags);
        }
      }
    }
  } 

  return 0;
}


// -------------------------------------------------------------------------

struct literal_list_struc *nonres_name_list_root = NULL;

// -------------------------------------------------------------------------
int build_nonresident_name_tree(void)
{
  char *curr_ptr;
  int literal_length;

  struct literal_list_struc *curr_nonres_name_list_node, *prev_nonres_name_list_node;

  nonres_name_list_root = NULL;

  curr_ptr = file_image + ne_header->OffStartNonResTab;

  while (-1)
  {
    literal_length = *((uint8_t*)curr_ptr);
    curr_ptr ++;

    if (! literal_length)
    {
      break;
    }

    if ((curr_nonres_name_list_node = malloc(sizeof(struct literal_list_struc))) == NULL)
    {
      printf("Failed to allocate memory for nonresident name list item\n");
      return -1;
    } else
    {
      curr_nonres_name_list_node->next = 0;

      if ((curr_nonres_name_list_node->literal_string = malloc(literal_length + 1)) == NULL)
      {
        printf("Failed to allocate memory for nonresident name literal string\n");
        return -1;
      } else
      {
        memcpy(curr_nonres_name_list_node->literal_string, curr_ptr, literal_length);
        curr_nonres_name_list_node->literal_string[literal_length] = 0;
      }

      curr_ptr += literal_length;
      curr_nonres_name_list_node->ordinal = *((uint16_t *)curr_ptr);
      curr_ptr += sizeof(uint16_t);

      // update pointers
      if (nonres_name_list_root == NULL)
      {
        nonres_name_list_root = curr_nonres_name_list_node;
      } else
      {
        prev_nonres_name_list_node->next = curr_nonres_name_list_node;
      }
      prev_nonres_name_list_node = curr_nonres_name_list_node;
    }
  }

  return 0;
}


// -------------------------------------------------------------------------
int display_nonresident_name_tree(void)
{
  struct literal_list_struc *nonres_name_list_node;

  if (nonres_name_list_root)  
  {
    printf("\nNonresident names by ordinal:\n");
    for (nonres_name_list_node = nonres_name_list_root; nonres_name_list_node; nonres_name_list_node = nonres_name_list_node->next)
    {
      printf("\t%s\t%04X\n", nonres_name_list_node->literal_string, nonres_name_list_node->ordinal);
    }
  }

  return 0;
}

// -------------------------------------------------------------------------
int delete_segment(int seg_ordinal)
{
  struct seg_list_entry *prev_seg_list_node, *curr_seg_list_node;
  struct reloc_list_entry *prev_reloc_list_node, *curr_reloc_list_node, *next_reloc_list_node;

  struct entry_bundle_list_struc *entry_bundle_list_node;
  struct entry_list_struc *entry_list_node;

  int segment_found;
  int i;

  if (seg_list_root == NULL)
  {
    printf("Segment list root is empty, exiting\n");
    return -1;
  }

  // find the segment to delete
  prev_seg_list_node = NULL;
  segment_found = 0;

  for (curr_seg_list_node = seg_list_root; curr_seg_list_node; curr_seg_list_node = curr_seg_list_node->next)
  {
    if (curr_seg_list_node->seg_ordinal == seg_ordinal)
    {
      segment_found = 1;
      break;
    }

    prev_seg_list_node = curr_seg_list_node;
  }

  if (! segment_found)
  {
    printf("Requested segment not found, exiting\n");
    return -1;
  }

  // segment is found

  // exclude current segment from the tree
  if (prev_seg_list_node != NULL)
  {
    // not the first segment
    prev_seg_list_node->next = curr_seg_list_node->next;
  } else
  {
    // the first segment
    seg_list_root = curr_seg_list_node->next;
  }

  // free all the things related to the segment
  if (curr_seg_list_node->seg_relocs)
  {
    for (curr_reloc_list_node = curr_seg_list_node->seg_relocs; 
      curr_reloc_list_node; curr_reloc_list_node = next_reloc_list_node)
    {
      next_reloc_list_node = curr_reloc_list_node->next;
      free(curr_reloc_list_node);
    }

    curr_seg_list_node->seg_relocs = NULL;
  }

  // free current segment entry
  free(curr_seg_list_node);

  // update segment ordinals and relocation trees
  i = 0;
  for (curr_seg_list_node = seg_list_root; curr_seg_list_node; curr_seg_list_node = curr_seg_list_node->next)
  {
    // update segment ordinal
    i ++;
    curr_seg_list_node->seg_ordinal = i;

    // scan and update relocation tree for current segment
    // relocations pointing to segment number > seg_ordinal must be decremented
    if (curr_seg_list_node->seg_relocs)
    {
      prev_reloc_list_node = NULL;

      for (curr_reloc_list_node = curr_seg_list_node->seg_relocs; 
        curr_reloc_list_node; curr_reloc_list_node = next_reloc_list_node)
      {
        if ((curr_reloc_list_node->rel_flags & REL_FLAGS_TARGET_MASK) == REL_FLAGS_INTERNALREF)
        {
          if ((curr_reloc_list_node->rel_target_segment & 0xFF) == seg_ordinal)
          {
            // delete relocation pointing to fixed segment to be removed
            if (prev_reloc_list_node)
            {
              // not the first node
              prev_reloc_list_node->next = curr_reloc_list_node->next;
            } else
            {
              // first node
              curr_seg_list_node->seg_relocs = curr_reloc_list_node->next;
            }

            next_reloc_list_node = curr_reloc_list_node->next;

            printf("Reloc INTERNALREF src %02X:%04X target %02X:%04X is removed...\n", 
              curr_seg_list_node->seg_ordinal, 
              curr_reloc_list_node->rel_src_offset,
              curr_reloc_list_node->rel_target_segment,  // = seg_ordinal
              curr_reloc_list_node->rel_target_offset);

            free (curr_reloc_list_node);

            continue;

          } else if (
            // if relocation points to segment after one being deleted
            ((curr_reloc_list_node->rel_target_segment & 0xFF) > seg_ordinal) &&
            ((curr_reloc_list_node->rel_target_segment & 0xFF) < 0xFF) )
          {
            // check if static segment with segment number > seg_ordinal
            // decrement the target segment number
          /*
            printf("Reloc INTERNALREF src %02X:%04X target %02X:%04X is modified...\n", 
              curr_seg_list_node->seg_ordinal, 
              curr_reloc_list_node->rel_src_offset,
              curr_reloc_list_node->rel_target_segment,  // = seg_ordinal
              curr_reloc_list_node->rel_target_offset);
          */
            curr_reloc_list_node->rel_target_segment --;
          }
        }

        next_reloc_list_node = curr_reloc_list_node->next;
        prev_reloc_list_node = curr_reloc_list_node;
      }
    }
  }



  if (entry_list_root)
  {
    for (entry_bundle_list_node = entry_list_root; 
      entry_bundle_list_node; entry_bundle_list_node = entry_bundle_list_node->next)
    {
      if (entry_bundle_list_node->segment_indicator == 0xFF)
      {
        // moveable
        if (entry_bundle_list_node->entries)
        {
          for (entry_list_node = entry_bundle_list_node->entries; entry_list_node; entry_list_node = entry_list_node->next)
          {
            if (entry_list_node->segment == seg_ordinal)
            {
              printf("movable bundle segment is equal to seg_ordinal, not implemented");
              continue;
            }
            if ((entry_list_node->segment > seg_ordinal) && (entry_list_node->segment < 0xFF))
            {
              printf("\tEntry %d ptr %02X:%04X flags %04X segment decremented...\n",
                entry_list_node->ordinal,
                entry_list_node->segment,
                entry_list_node->offset,
                entry_list_node->flags);

              // decrement segment
              entry_list_node->segment --;

            }
          }
        }
      } else
      {
        // static
        if (entry_bundle_list_node->segment_indicator == seg_ordinal)
        {
          printf("static bundle segment indicator is equal to seg_ordinal, not implemented");
          continue;
        }

        if ((entry_bundle_list_node->segment_indicator > seg_ordinal) &&
          (entry_bundle_list_node->segment_indicator < 0xFF))
        {
          // decrement segment indicator
          entry_bundle_list_node->segment_indicator --;
        }
      }
    }
  } else
  {
    printf("Entry list root is empty...\n");
  } 


  seg_count --;

  return 0;
}

                    
// -------------------------------------------------------------------------
int update_pascal_seg_indexes(void)
{
  struct seg_list_entry *seg_list_node;

  if (seg_list_root != NULL)
  {
    for (seg_list_node = seg_list_root; seg_list_node; seg_list_node = seg_list_node->next)
    {
      if (!(seg_list_node->seg_flags & SEG_FLAGS_DATA))
      {
        // pascal code seg
        *((uint16_t *)seg_list_node->seg_data) = seg_list_node->seg_ordinal;
      }
    }

    return 0;
  } else
  {
    printf("Segment list root is empty...\n");
    return -1;
  }
}

// -------------------------------------------------------------------------

int build_output(char *output_file_name)
{
  // of common use
  char *new_file_image = NULL;
  int new_file_length;
  int dos_hdr_length;
  struct NE_header_struc *new_ne_header;
  char *new_ne_image;
  char *curr_ptr;
  int i;
  struct literal_list_struc *literal_list_node;
  int literal_length;
  FILE *output_file;

  int sort_done_flag;

  // to build segment list
  struct seg_list_entry *seg_list_node;
  struct reloc_list_entry *reloc_list_node;
  struct NE_seg_struc seg_entry;
  struct NE_reloc_entry_struc reloc_entry;
  int seg_align_mask;
  int segment_base_offset;
  int seg_table_index;

  // to build resource table
  char *rsrc_table_base;
  struct rsrc_type_struc rsrc_type;
  struct rsrc_entry_struc rsrc_entry;
  struct rsrc_type_list_struc *rsrc_type_list_node;
  struct rsrc_entry_list_struc *rsrc_entry_list_node;
  int rsrc_base_offset;
  int rsrc_align_mask;
  char *rsrc_table_ptr;

  struct rsrc_entry_list_struc ** rsrc_entry_array = NULL;
  int rsrc_entry_count;


  // to build resource literal table
  struct literal_list_struc ** rsrc_literal_array = NULL;
  int rsrc_literal_array_size;

  // to build module reference table
  int mod_ref_table_size;
  uint16_t *mod_ref_table_ptr;
  char *import_name_table_ptr;

  // to build entry table
  struct entry_bundle_list_struc *entry_bundle_list_node;
  struct entry_list_struc *entry_list_node;
  char *entry_table_ptr;



  if((new_file_image = (char *)malloc(file_length * 2)) == NULL) // why not * 2?
  {
    printf("**Error**: failed to allocate memory for output file\n");
    return -2;
  }

  memset(new_file_image, 0, file_length * 2);

  curr_ptr = new_file_image;

  dos_hdr_length = * (unsigned short *)(file_image + 0x3C);
  memcpy(curr_ptr, file_image, dos_hdr_length); // copy dos header
  curr_ptr += dos_hdr_length;

  memcpy(curr_ptr, file_image + dos_hdr_length, sizeof(struct NE_header_struc)); // copy NE header

  new_ne_image = curr_ptr;
  new_ne_header = (struct NE_header_struc *)curr_ptr;

  new_ne_header->FileLoadCRC = 0;
  new_ne_header->EntryTableOffset = 0;
  new_ne_header->EntryTableLength = 0;
  new_ne_header->SegCount = 0;
  new_ne_header->ModRefs = 0;
  new_ne_header->ResidNamTable = 0;
  new_ne_header->ModRefTable = 0;
  new_ne_header->ImportNameTable = 0;
  new_ne_header->OffStartNonResTab = 0;
  new_ne_header->nResTabEntries = 0; // and will be zero 
  new_ne_header->ResTableOffset = 0;

  curr_ptr += sizeof(struct NE_header_struc);


//  curr_ptr += 0x10; // just for fun, check if all things are working


  // build segment table, pass 1
  new_ne_header->SegTableOffset = curr_ptr - new_ne_image;
  new_ne_header->SegCount = seg_count;
  // leave a space for segment entries
  curr_ptr += sizeof(struct NE_seg_struc) * seg_count;


  // build resource table
  new_ne_header->ResTableOffset = curr_ptr - new_ne_image;
  rsrc_table_base = curr_ptr;

  // store alignment
  *((uint16_t *)curr_ptr) = rsrc_alignment;
  curr_ptr += sizeof(uint16_t);

  if (rsrc_list_root)
  {
    rsrc_literal_array_size = 0;

    // store entries
    for(rsrc_type_list_node = rsrc_list_root; rsrc_type_list_node; rsrc_type_list_node = rsrc_type_list_node->next)
    {
      rsrc_type = rsrc_type_list_node->raw_entry;
      // store resource type structure
      memcpy(curr_ptr, &rsrc_type, sizeof(rsrc_type));
      curr_ptr += sizeof(rsrc_type);

      for (rsrc_entry_list_node = rsrc_type_list_node->rsrc_entries; 
        rsrc_entry_list_node; rsrc_entry_list_node = rsrc_entry_list_node->next)
      {
        rsrc_entry = rsrc_entry_list_node->raw_entry;
        // store resource entry structure
        memcpy(curr_ptr, &rsrc_entry, sizeof(rsrc_entry));
        curr_ptr += sizeof(rsrc_entry);

        // update number of literals
        if (rsrc_entry_list_node->literal_entry)
        {
          rsrc_literal_array_size ++;
        }
      }
    }
    *((uint16_t *)curr_ptr) = 0; // store 0 type id to indicate the end of list
    curr_ptr += sizeof(uint16_t);

    // allocate array of literals to sort
    if ((rsrc_literal_array = malloc(sizeof(struct literal_list_struc *) * rsrc_literal_array_size)) == NULL)
    {
      printf("Failed to allocate memory for resource literal array\n");
      return -1;
    } else
    {
      // fill array
      i = 0;
      for(rsrc_type_list_node = rsrc_list_root; rsrc_type_list_node; rsrc_type_list_node = rsrc_type_list_node->next)
      {
        for (rsrc_entry_list_node = rsrc_type_list_node->rsrc_entries; 
          rsrc_entry_list_node; rsrc_entry_list_node = rsrc_entry_list_node->next)
        {
          if (rsrc_entry_list_node->literal_entry)
          {
            rsrc_literal_array[i] = rsrc_entry_list_node->literal_entry;
            i ++;
          }
        }
      }

      // sort array sequence as it was in the source file
      sort_done_flag = 0;
      while (sort_done_flag == 0)
      {
        sort_done_flag = 1;
        for (i = 0; i < rsrc_literal_array_size - 1; i ++)
        {
          if (rsrc_literal_array[i]->offset > rsrc_literal_array[i + 1]->offset)
          {
            literal_list_node = rsrc_literal_array[i];
            rsrc_literal_array[i] = rsrc_literal_array[i + 1];
            rsrc_literal_array[i + 1] = literal_list_node;
            sort_done_flag = 0;
          }
        } 
      }
    }

    // do store the literals in resource table
    for (i = 0; i < rsrc_literal_array_size; i ++)
    {
      // update offset for new resource table
      rsrc_literal_array[i]->offset = curr_ptr - rsrc_table_base;
      // store string length
      literal_length = strlen(rsrc_literal_array[i]->literal_string);
      *((uint8_t *)curr_ptr) = literal_length;
      curr_ptr ++;
      // store literal string
      memcpy(curr_ptr, rsrc_literal_array[i]->literal_string, literal_length);
      curr_ptr += literal_length;
    }

    free(rsrc_literal_array); rsrc_literal_array = NULL;

  } else
  {
    // it is not forbidden to have empty resource list, 
    // but we may have forgot to parse the resource tree
    printf("Resource list root is empty...\n");

    *((uint16_t *)curr_ptr) = 0; // store 0 type id to indicate the end of list
    curr_ptr += sizeof(uint16_t);
  }


  // build resident name table
  if (res_name_list_root)
  {
    new_ne_header->ResidNamTable = curr_ptr - new_ne_image;
    for (literal_list_node = res_name_list_root; literal_list_node; literal_list_node = literal_list_node->next)
    {
      // store string length
      literal_length = strlen(literal_list_node->literal_string);
      *((uint8_t *)curr_ptr) = literal_length;
      curr_ptr ++;

      // store literal string
      memcpy(curr_ptr, literal_list_node->literal_string, literal_length);
      curr_ptr += literal_length;

      // store ordinal
      *((uint16_t *)curr_ptr) = literal_list_node->ordinal;
      curr_ptr += sizeof(uint16_t);
    }

    // store terminator
    *((uint8_t *)curr_ptr) = 0;
    curr_ptr += sizeof(uint8_t);
  } else
  {
    printf("Resident name list root is empty...\n");
  }


  // build module reference table
  new_ne_header->ModRefTable = curr_ptr - new_ne_image;
  mod_ref_table_ptr = (uint16_t *)curr_ptr;

  if (module_list_root)
  {
    // determine size of module reference table
    i = 0;
    for (literal_list_node = module_list_root; literal_list_node; literal_list_node = literal_list_node->next)
    {
      i ++;
    }

    mod_ref_table_size = i;
    curr_ptr += mod_ref_table_size * sizeof(uint16_t);

    new_ne_header->ImportNameTable = curr_ptr - new_ne_image;
    import_name_table_ptr = curr_ptr;

    // store leading zero in import table
    *((uint8_t *)curr_ptr) = 0;
    curr_ptr += sizeof(uint8_t);

    // store literals
    i = 0;
    for (literal_list_node = module_list_root; literal_list_node; literal_list_node = literal_list_node->next)
    {
      // store offset in module reference table
      mod_ref_table_ptr[i] = curr_ptr - import_name_table_ptr;

      // store string length
      literal_length = strlen(literal_list_node->literal_string);
      *((uint8_t *)curr_ptr) = literal_length;
      curr_ptr ++;

      // store literal string
      memcpy(curr_ptr, literal_list_node->literal_string, literal_length);
      curr_ptr += literal_length;

      i ++;
    }

    new_ne_header->ModRefs = mod_ref_table_size;
  }


  // build entry table
  new_ne_header->EntryTableOffset = curr_ptr - new_ne_image;
  entry_table_ptr = curr_ptr;

  if (entry_list_root)
  {
    for (entry_bundle_list_node = entry_list_root; 
      entry_bundle_list_node; entry_bundle_list_node = entry_bundle_list_node->next)
    {
      // store entry count
      *((uint8_t *)curr_ptr) = entry_bundle_list_node->entry_count;
      curr_ptr ++;
      // store segment indicator
      *((uint8_t *)curr_ptr) = entry_bundle_list_node->segment_indicator;
      curr_ptr ++;
      if (entry_bundle_list_node->entries)
      {
        for (entry_list_node = entry_bundle_list_node->entries; entry_list_node; entry_list_node = entry_list_node->next)
        {
          if (entry_list_node->size == 6)
          {
            // movable
            *((uint8_t *)curr_ptr) = entry_list_node->flags;
            *((uint16_t *)(curr_ptr + 1)) = 0x3FCD; // int 3F
            *((uint8_t *)(curr_ptr + 3)) = entry_list_node->segment;
            *((uint16_t *)(curr_ptr + 4)) = entry_list_node->offset;
            curr_ptr += 6;
          } else
          {
            // fixed
            *((uint8_t *)curr_ptr) = entry_list_node->flags;
            *((uint16_t *)(curr_ptr + 1)) = entry_list_node->offset;
            curr_ptr += 3;
          }
        }
      }
    }

    // store terminators
    curr_ptr += 4; // memory is already zero
    // store entry table length
    new_ne_header->EntryTableLength = curr_ptr - entry_table_ptr;

  } else
  {
    printf("Entries list root is empty...\n");
  }


  // build nonresident name table
  new_ne_header->OffStartNonResTab = curr_ptr - new_file_image; // pointers from the very beginning of the file

  if (nonres_name_list_root)  
  {
    for (literal_list_node = nonres_name_list_root; literal_list_node; literal_list_node =literal_list_node->next)
    {
      // store string length
      literal_length = strlen(literal_list_node->literal_string);
      *((uint8_t *)curr_ptr) = literal_length;
      curr_ptr ++;

      // store literal string
      memcpy(curr_ptr, literal_list_node->literal_string, literal_length);
      curr_ptr += literal_length;

      // store ordinal
      *((uint16_t *)curr_ptr) = literal_list_node->ordinal;
      curr_ptr += sizeof(uint16_t);
    }

    // store terminator
    *((uint8_t *)curr_ptr) = 0;
    curr_ptr += sizeof(uint8_t);
  }


  // store segment data
  if (seg_list_root)
  {
    seg_align_mask = (1 << new_ne_header->FileAlnSzShftCnt) - 1;

    seg_table_index = 0; // segment table index
    for (seg_list_node = seg_list_root; seg_list_node; seg_list_node = seg_list_node->next)
    {
      segment_base_offset = ((curr_ptr - new_file_image) + seg_align_mask) & (~seg_align_mask);
      curr_ptr = new_file_image + segment_base_offset;

      seg_entry.seg_offset_in_file = segment_base_offset >> new_ne_header->FileAlnSzShftCnt; 
      seg_entry.seg_len = (seg_list_node->seg_length != 0x10000) ? seg_list_node->seg_length : 0;
      seg_entry.seg_flags = seg_list_node->seg_flags;
      seg_entry.seg_alloc_size = (seg_list_node->seg_alloc_length != 0x10000) ? seg_list_node->seg_alloc_length : 0;

      // store segment table entry
      memcpy(new_ne_image + new_ne_header->SegTableOffset + seg_table_index * sizeof(seg_entry), 
        &seg_entry, sizeof(seg_entry));
      seg_table_index++;

      // copy segment data
      memcpy(curr_ptr, seg_list_node->seg_data, seg_list_node->seg_length);
      curr_ptr += seg_list_node->seg_length;

      // store relocations if available
      if (seg_list_node->seg_relocs)
      {
        // calculate a number of relocations
        i = 0;
        for (reloc_list_node = seg_list_node->seg_relocs; reloc_list_node; reloc_list_node = reloc_list_node->next)
        {
          i++;
        }

        seg_list_node->seg_reloc_count = i;

        // store the number of relocations
        *((uint16_t *)curr_ptr) = seg_list_node->seg_reloc_count;
        curr_ptr += sizeof(uint16_t);

        // store relocations
        for (reloc_list_node = seg_list_node->seg_relocs; reloc_list_node; reloc_list_node = reloc_list_node->next)
        {


//          reloc_entry = reloc_list_node->raw_entry;
          // encoding raw part
          reloc_entry.rel_src_type = reloc_list_node->rel_src_type;
          reloc_entry.rel_flags = reloc_list_node->rel_flags;
          reloc_entry.rel_src_offset = reloc_list_node->rel_src_offset;

          switch(reloc_list_node->rel_flags & REL_FLAGS_TARGET_MASK)
          {
            case REL_FLAGS_INTERNALREF:
              reloc_entry.rel_param_1 = reloc_list_node->rel_target_segment & 0xFF;

              if ((reloc_list_node->rel_target_segment & 0xFF) == 0xFF)
              {
                // movable segment
                reloc_entry.rel_param_2 = reloc_list_node->rel_target_entry_ordinal;
              } else
              {
                reloc_entry.rel_param_2 = reloc_list_node->rel_target_offset;
              }
              break;

            case REL_FLAGS_IMPORTORDINAL:
              reloc_entry.rel_param_1 = reloc_list_node->rel_module_ordinal;
              reloc_entry.rel_param_2 = reloc_list_node->rel_proc_ordinal;
              break;

            case REL_FLAGS_IMPORTNAME:
              printf("Relocation IMPORTNAME: not implemented\n");
              reloc_entry.rel_param_1 = reloc_list_node->rel_module_ordinal;
              reloc_entry.rel_param_2 = reloc_list_node->rel_proc_name_ptr;   // ptr may have been changed
              break;

            case REL_FLAGS_OSFIXUP:
              reloc_entry.rel_param_1 = reloc_list_node->rel_osfixup;
              reloc_entry.rel_param_2 = 0;

          }



          memcpy(curr_ptr, &reloc_entry, sizeof(reloc_entry));
          curr_ptr += sizeof(reloc_entry);
        }

      } else
      {
        curr_ptr += sizeof(uint16_t); // leave zero relocs
      }
    }
  } else
  {
    printf("Segment list root is empty, exiting\n");
    return -1;
  }


  // store resource data
  if (rsrc_list_root)
  {
    rsrc_align_mask = (1 << rsrc_alignment) - 1;
    rsrc_base_offset = ((curr_ptr - new_file_image) + rsrc_align_mask) & (~rsrc_align_mask);

    i = 0;
    for(rsrc_type_list_node = rsrc_list_root; rsrc_type_list_node; rsrc_type_list_node = rsrc_type_list_node->next)
    {
      if (rsrc_type_list_node->rsrc_entries)
      {
        for (rsrc_entry_list_node = rsrc_type_list_node->rsrc_entries; 
          rsrc_entry_list_node; rsrc_entry_list_node = rsrc_entry_list_node->next)
        {
          i++;
        }
      }
    }

    rsrc_entry_count = i;

    if ((rsrc_entry_array = malloc(sizeof(struct rsrc_entry_list_struc *) * rsrc_entry_count)) == NULL)
    {
      printf("Failed to allocate memory for resource entries array\n");
      return -1;
    } else
    {
      i = 0;
      for(rsrc_type_list_node = rsrc_list_root; rsrc_type_list_node; rsrc_type_list_node = rsrc_type_list_node->next)
      {
        if (rsrc_type_list_node->rsrc_entries)
        {
          for (rsrc_entry_list_node = rsrc_type_list_node->rsrc_entries; 
            rsrc_entry_list_node; rsrc_entry_list_node = rsrc_entry_list_node->next)
          {
            rsrc_entry_array[i] = rsrc_entry_list_node;
            i++;
          }
        }
      }

      // sort resources as they appear in the source file
      sort_done_flag = 0;
      while (! sort_done_flag)
      {
        sort_done_flag = 1;
        for (i = 0; i < rsrc_entry_count - 1; i++)
        {
          if (rsrc_entry_array[i]->rsrc_offset > rsrc_entry_array[i + 1]->rsrc_offset)
          {
            rsrc_entry_list_node = rsrc_entry_array[i];
            rsrc_entry_array[i] = rsrc_entry_array[i + 1];
            rsrc_entry_array[i + 1] = rsrc_entry_list_node;
            sort_done_flag = 0;
          }
        }
      }

      // copy all the resource data
      for (i = 0; i < rsrc_entry_count; i++)
      {
        rsrc_entry_array[i]->rsrc_length = (rsrc_entry_array[i]->rsrc_length + rsrc_align_mask) & (~rsrc_align_mask);
        rsrc_entry_array[i]->rsrc_offset = rsrc_base_offset;
        rsrc_base_offset += rsrc_entry_array[i]->rsrc_length;

        memcpy(new_file_image + rsrc_entry_array[i]->rsrc_offset, rsrc_entry_array[i]->rsrc_data, rsrc_entry_array[i]->rsrc_length);
      }

      // fill resource table
      rsrc_table_ptr = rsrc_table_base + 2; // skip alignment

      for(rsrc_type_list_node = rsrc_list_root; rsrc_type_list_node; rsrc_type_list_node = rsrc_type_list_node->next)
      {
        rsrc_type = rsrc_type_list_node->raw_entry;
    
        // store resource type structure at rsrc_table_ptr pointing to the header resource table
        memcpy(rsrc_table_ptr, &rsrc_type, sizeof(rsrc_type));
        rsrc_table_ptr += sizeof(rsrc_type);

        if (rsrc_type_list_node->rsrc_entries)
        {
          for (rsrc_entry_list_node = rsrc_type_list_node->rsrc_entries; 
            rsrc_entry_list_node; rsrc_entry_list_node = rsrc_entry_list_node->next)
          {
            rsrc_entry = rsrc_entry_list_node->raw_entry;
            rsrc_entry.rsrc_offset = rsrc_entry_list_node->rsrc_offset >> rsrc_alignment;
    
            // store offset of literal entry if available
            if (rsrc_entry_list_node->literal_entry)
            {
              rsrc_entry.rsrc_id = rsrc_entry_list_node->literal_entry->offset;
            }

            // store resource entry structure at rsrc_table_ptr
            memcpy(rsrc_table_ptr, &rsrc_entry, sizeof(rsrc_entry));
            rsrc_table_ptr += sizeof(rsrc_entry);
          }
        }
      }

      new_file_length = rsrc_entry_array[rsrc_entry_count - 1]->rsrc_offset + rsrc_entry_array[rsrc_entry_count - 1]->rsrc_length;
    }
  } else
  {
    printf("Resource list root is empty, exiting\n");
    return -1;
  }


  if ((output_file = fopen(output_file_name, "wb")) == NULL)
  {
    printf("Failed to open output file, exiting\n");
    return -1;
  } else
  {
    fwrite(new_file_image, sizeof(char), new_file_length, output_file);

    fclose(output_file);
  }

  free(new_file_image);

  return 0;
}

// -------------------------------------------------------------------------

int main(int argc, char **argv)
{
  int segment;
  char buffer[20];

/*
  if (argc != 2)
  {
    printf("NE dump v. 0.1\n");
    return 0;
  }

  load_image(argv[1]);
  parse_ne_header();
*/

/*
  load_image("SOFTSP_D.EXE");
  parse_ne_header();

  build_segment_tree(1);
  build_resource_tree();
  build_resident_name_tree();
  build_module_tree();
  build_entry_tree();
  build_nonresident_name_tree();

  build_output("SOFTSP_D.EXE.out");
*/

  load_image("SOFTS_c4.exe");
  parse_ne_header();

  build_segment_tree(0);
  build_resource_tree();
  build_resident_name_tree();
  build_module_tree();
  build_entry_tree();
  build_nonresident_name_tree();

  for (segment = 1; segment < 0x2F; segment ++)
  {
    sprintf(buffer, "FIXUP_%02X.BIN", segment);
    load_relocs_tree(segment, buffer);
  }

//  display_segment_tree();

  delete_resource_by_literal("GHITA");
  delete_segment(0x30);
  delete_segment(0x2F);

  export_segment("seg_1b.bin", 0x1b);

  delete_segment(0x1B);

  delete_reloc_src(20, 0x4540);
  delete_reloc_src(20, 0x454B);

  update_pascal_seg_indexes();


/*
delete_reloc_src(0x01, 0x00F1);
delete_reloc_src(0x02, 0x0E3E);
delete_reloc_src(0x02, 0x0EED);
delete_reloc_src(0x02, 0x0F9C);
delete_reloc_src(0x02, 0x109C);
delete_reloc_src(0x02, 0x11D6);
delete_reloc_src(0x02, 0x18AE);
delete_reloc_src(0x02, 0x1CA3);
delete_reloc_src(0x02, 0x36A1);
delete_reloc_src(0x03, 0x10FE);
delete_reloc_src(0x03, 0x110C);
delete_reloc_src(0x07, 0x2F9E);
delete_reloc_src(0x11, 0x11D2);
delete_reloc_src(0x11, 0x6EDA);
delete_reloc_src(0x11, 0x72AB);
delete_reloc_src(0x14, 0x4537);
delete_reloc_src(0x14, 0x4540);
delete_reloc_src(0x14, 0x8D5D);
delete_reloc_src(0x14, 0x986B);
*/



  ne_header->EntryPoint = 0x10091; // 01:0091
  ne_header->AutoDataSegIndex = 0x2D;
  ne_header->InitStack = 0x2D0000; // 2D:0000  //SS:SP inital stack pointer, SS is index into segment table
//  ne_header->AutoDataSegIndex = 0x2E;
//  ne_header->InitStack = 0x2E0000; // 2E:0000  //SS:SP inital stack pointer, SS is index into segment table


  build_output("softs_c4.exe.new");

//  display_segment_tree();

//  display_resource_tree();

/*
  build_resource_tree();
  display_resource_tree();
*/
/*
  build_resident_name_tree();
  display_resident_name_tree();
*/
/*
  build_module_tree();
  display_module_tree();
*/
/*
  build_entry_tree();
  display_entry_tree();
*/
/*
  build_nonresident_name_tree();
  display_nonresident_name_tree();
*/
/*
  parse_segments();
  parse_name_table();
  parse_resources();
  parse_module_entries();
  parse_entries();
*/
  return 0;

}
